import logging
from typing import Any, Callable, Dict, List, Mapping, Optional, Set, Tuple, Type

from flask import Flask, Request, request  # type: ignore
from flask.views import MethodView
from werkzeug.datastructures import EnvironHeaders, ImmutableMultiDict

from pait.app.base import BaseAppHelper
from pait.core import pait as _pait
from pait.g import pait_data
from pait.model import PaitCoreModel, PaitResponseModel, PaitStatus
from pait.util import LazyProperty


class AppHelper(BaseAppHelper):
    RequestType = Request
    FormType = ImmutableMultiDict
    FileType = Request.files
    HeaderType = EnvironHeaders
    app_name = "flask"

    def __init__(self, class_: Any, args: Tuple[Any, ...], kwargs: Mapping[str, Any]):
        super().__init__(class_, args, kwargs)

        self.request = request

    def body(self) -> dict:
        return request.json

    def cookie(self) -> dict:
        return request.cookies

    def file(self) -> Request.files:  # type: ignore
        return request.files

    def form(self) -> Request.form:  # type: ignore
        return request.form

    def header(self) -> EnvironHeaders:
        return request.headers

    def path(self) -> Mapping[str, Any]:
        return self.request_kwargs

    def query(self) -> Dict[str, Any]:
        return request.args

    @LazyProperty()
    def multiform(self) -> Dict[str, List[Any]]:
        return {key: request.form.getlist(key) for key, _ in request.form.items()}

    @LazyProperty()
    def multiquery(self) -> Dict[str, List[Any]]:
        return {key: request.args.getlist(key) for key, _ in request.args.items()}


def load_app(app: Flask) -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`"""
    _pait_data: Dict[str, PaitCoreModel] = {}
    for route in app.url_map.iter_rules():
        path: str = route.rule
        method_set: Set[str] = route.methods
        route_name: str = route.endpoint
        endpoint: Callable = app.view_functions[route_name]
        pait_id: Optional[str] = getattr(endpoint, "_pait_id", None)
        if not pait_id:
            if route_name == "static":
                continue
            view_class_endpoint = getattr(endpoint, "view_class", None)
            if not view_class_endpoint or not issubclass(view_class_endpoint, MethodView):
                logging.warning(f"loan path:{path} fail, endpoint:{endpoint} not `view_class` attributes")
                continue
            for method in view_class_endpoint.methods:
                method = method.lower()
                method_set = {method}
                endpoint = getattr(view_class_endpoint, method, None)
                if not endpoint:
                    continue
                pait_id = getattr(endpoint, "_pait_id", None)
                if not pait_id:
                    continue
                pait_data.add_route_info(AppHelper.app_name, pait_id, path, method_set, f"{route_name}.{method}")
                _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
        else:
            pait_data.add_route_info(AppHelper.app_name, pait_id, path, method_set, route_name)
            _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    return _pait_data


def pait(
    author: Optional[Tuple[str]] = None,
    desc: Optional[str] = None,
    status: Optional[PaitStatus] = None,
    group: Optional[str] = None,
    tag: Optional[Tuple[str, ...]] = None,
    response_model_list: List[Type[PaitResponseModel]] = None,
) -> Callable:
    """Help flask provide parameter checks and type conversions for each routing function/cbv class"""
    return _pait(
        AppHelper,
        author=author,
        desc=desc,
        status=status,
        group=group,
        tag=tag,
        response_model_list=response_model_list,
    )



import copy
import json
from typing import Any, Dict, List, Optional, Tuple, Type

import yaml
from pydantic import BaseModel, Field, HttpUrl
from pydantic.fields import Undefined

from pait import field as pait_field
from pait.api_doc.base_parse import PaitBaseParse
from pait.model import PaitCoreModel, PaitResponseModel, PaitStatus
from pait.util import create_pydantic_model

__all__ = ["PaitOpenApi"]


class _OpenApiInfoModel(BaseModel):
    """open api info column model"""

    class _Contact(BaseModel):
        name: str
        url: str
        email: str

    class _License(BaseModel):
        name: str
        url: str

    title: str = Field("Pait Open Api")
    description: str = Field(None)
    version: str = Field("0.0.1")
    contact: _Contact = Field(None)
    license: _License = Field(None)


class _OpenApiTagModel(BaseModel):
    """openapi tag column model"""

    class _ExternalDocs(BaseModel):
        url: HttpUrl

    name: str
    description: str = Field(None)
    externalDocs: str = Field(None)


class _OpenApiServerModel(BaseModel):
    """openapi server column model"""

    url: HttpUrl = Field("http://127.0.0.1")
    description: str = Field(None)


class PaitOpenApi(PaitBaseParse):
    def __init__(
        self,
        pait_dict: Dict[str, PaitCoreModel],
        title: Optional[str] = None,
        open_api_info: Optional[Dict[str, Any]] = None,
        open_api_tag_list: Optional[List[Dict[str, Any]]] = None,
        open_api_server_list: Optional[List[Dict[str, Any]]] = None,
        # default_response: Optional[...] = None,  # TODO
        type_: str = "json",
        filename: Optional[str] = None,
    ):
        super().__init__(pait_dict)
        self._header_keyword_dict: Dict[str, str] = {
            "Content-Type": "requestBody.content.<media-type>",
            "Accept": "responses.<code>.content.<media-type>",
            "Authorization": " security",
        }

        if not open_api_info:
            open_api_info = _OpenApiInfoModel(title=title).dict(exclude_none=True)
        else:
            open_api_info = _OpenApiInfoModel(**open_api_info).dict()

        if not open_api_server_list:
            open_api_server_list = [_OpenApiServerModel().dict(exclude_none=True)]
        else:
            temp_open_api_server_list: List[Dict[str, Any]] = []
            for open_api_server in open_api_server_list:
                temp_open_api_server_list.append(_OpenApiServerModel(**open_api_server).dict(exclude_none=True))
            open_api_server_list = temp_open_api_server_list

        if not open_api_tag_list:
            open_api_tag_list = []
        else:
            temp_open_api_tag_list: List[Dict[str, Any]] = []
            for open_api_tag in open_api_tag_list:
                temp_open_api_tag_list.append(_OpenApiTagModel(**open_api_tag).dict(exclude_none=True))
            open_api_tag_list = temp_open_api_tag_list

        open_api_dict: Dict[str, Any] = {
            "openapi": "3.0.0",
            "info": open_api_info,
            "servers": open_api_server_list,
            "tags": open_api_tag_list,
            "paths": {},
            "components": {"schemas": {}},
            # TODO
            # "security": {},
            # "externalDocs": {}
        }
        self.parse_data_2_openapi(open_api_dict)
        if type_ == "json":
            pait_json: str = json.dumps(open_api_dict)
            self.output(filename, pait_json, ".json")
        elif type_ == "yaml":
            pait_yaml: str = yaml.dump(open_api_dict, sort_keys=False)
            self.output(filename, pait_yaml, ".yaml")

    def replace_pydantic_definitions(
        self, schema: dict, path: str, open_api_dict: Dict[str, Any], parent_schema: Optional[dict] = None
    ) -> None:
        if not parent_schema:
            parent_schema = schema
        for key, value in schema.items():
            if key == "$ref" and not value.startswith("#/components"):
                index: int = value.rfind("/") + 1
                model_key: str = value[index:]
                schema[key] = f"#/components/schemas/{model_key}"
                open_api_dict["components"]["schemas"][model_key] = parent_schema["definitions"][model_key]
            if type(value) is dict:
                self.replace_pydantic_definitions(value, path, open_api_dict, parent_schema)

    @staticmethod
    def field_2_request_body(media_type: str, method_dict: dict, field_dict_list: List[dict]) -> None:
        openapi_request_body_dict: dict = method_dict.setdefault("requestBody", {"content": {}})

        annotation_dict: Dict[str, Tuple[Type, Any]] = {
            field_dict["raw"]["param_name"]: (field_dict["raw"]["annotation"], field_dict["raw"]["field"])
            for field_dict in field_dict_list
        }
        _pydantic_model: Type[BaseModel] = create_pydantic_model(annotation_dict)
        openapi_request_body_dict["content"].update({media_type: {"schema": _pydantic_model.schema()}})

    def parse_data_2_openapi(self, open_api_dict: Dict[str, Any]) -> None:
        for group, pait_model_list in self._group_pait_dict.items():
            for pait_model in pait_model_list:
                path: str = pait_model.path
                openapi_path_dict: dict = open_api_dict["paths"].setdefault(path, {})
                method_list: list = pait_model.method_list
                for method in method_list:
                    openapi_method_dict: dict = openapi_path_dict.setdefault(method.lower(), {})
                    if pait_model.tag:
                        openapi_method_dict["tags"] = list(pait_model.tag)
                        for tag in pait_model.tag:
                            if tag not in {tag_dict["name"] for tag_dict in open_api_dict["tags"]}:
                                open_api_dict["tags"].append(
                                    {
                                        "name": tag,
                                        "description": "",
                                    }
                                )
                    if pait_model.status in (
                        PaitStatus.abnormal,
                        PaitStatus.maintenance,
                        PaitStatus.archive,
                        PaitStatus.abandoned,
                    ):
                        openapi_method_dict["deprecated"] = True
                    openapi_method_dict["summary"] = pait_model.desc
                    openapi_method_dict["operationId"] = f"{method}.{pait_model.operation_id}"
                    openapi_parameters_list: list = openapi_method_dict.setdefault("parameters", [])
                    openapi_response_dict: dict = openapi_method_dict.setdefault("responses", {})
                    all_field_dict: Dict[str, List[Dict[str, Any]]] = self._parse_func_param_to_field_dict(
                        pait_model.func
                    )

                    for field, field_dict_list in all_field_dict.items():
                        if field in (
                            pait_field.Cookie.__name__.lower(),
                            pait_field.Header.__name__.lower(),
                            pait_field.Path.__name__.lower(),
                            pait_field.Query.__name__.lower(),
                        ):
                            for field_dict in field_dict_list:
                                param_name: str = field_dict["raw"]["param_name"]
                                if field == pait_field.Header.__name__.lower():
                                    param_name = self._header_keyword_dict.get(param_name, param_name)

                                # TODO support example
                                openapi_parameters_list.append(
                                    {
                                        "name": param_name,
                                        "in": field.lower(),
                                        "required": field_dict["default"] is Undefined,
                                        "description": field_dict["description"],
                                        "schema": field_dict["raw"]["schema"],
                                    }
                                )
                        elif field == pait_field.Body.__name__.lower():
                            # support args BodyField
                            self.field_2_request_body("application/json", openapi_method_dict, field_dict_list)
                        elif field == pait_field.Form.__name__.lower():
                            # support args FormField
                            self.field_2_request_body(
                                "application/x-www-form-urlencoded", openapi_method_dict, field_dict_list
                            )
                        else:
                            # TODO
                            pass

                    if pait_model.response_model_list:
                        response_schema_dict: Dict[tuple, List[Dict[str, str]]] = {}
                        for resp_model_class in pait_model.response_model_list:
                            resp_model: PaitResponseModel = resp_model_class()
                            schema_dict: dict = {}
                            if resp_model.response_data:
                                schema_dict = resp_model.response_data.schema()

                            # fix del schema dict
                            schema_dict = copy.deepcopy(schema_dict)
                            path = f"#/components/schemas/{schema_dict['title']}"
                            self.replace_pydantic_definitions(schema_dict, path, open_api_dict)
                            if "definitions" in schema_dict:
                                del schema_dict["definitions"]
                            for _status_code in resp_model.status_code:
                                key: tuple = (_status_code, resp_model.media_type)
                                ref_dict: dict = {"$ref": path}
                                if key in response_schema_dict:
                                    response_schema_dict[key].append(ref_dict)
                                else:
                                    response_schema_dict[key] = [ref_dict]
                                if _status_code in openapi_response_dict:
                                    openapi_response_dict[_status_code]["description"] += f"|{resp_model.description}"
                                else:
                                    openapi_response_dict[_status_code] = {"description": resp_model.description}
                                open_api_dict["components"]["schemas"].update({schema_dict["title"]: schema_dict})
                        # mutli response support
                        # only response example see https://swagger.io/docs/specification/describing-responses/   FAQ
                        for key_tuple, path_list in response_schema_dict.items():
                            status_code, media_type = key_tuple
                            if len(path_list) == 1:
                                ref_dict = path_list[0]
                            else:
                                ref_dict = {"oneOf": path_list}
                            openapi_response_dict[status_code]["content"] = {media_type: {"schema": ref_dict}}



import asyncio
import inspect
import logging
from types import ModuleType
from typing import Any, Callable, Coroutine, Dict, List, Mapping, NoReturn, Optional, Tuple, Type, Union, get_type_hints

from pydantic import BaseModel, fields

from pait import field
from pait.app.base import BaseAppHelper
from pait.exceptions import NotFoundFieldError, NotFoundValueError, PaitBaseException
from pait.field import BaseField
from pait.model import PaitBaseModel
from pait.util import FuncSig, create_pydantic_model, get_func_sig, get_parameter_list_from_class


def raise_and_tip(_object: Any, exception: "Exception", parameter: Optional[inspect.Parameter] = None) -> NoReturn:
    """Help users understand which parameter is wrong"""
    if parameter:
        param_value: BaseField = parameter.default
        annotation: Type[BaseModel] = parameter.annotation
        param_name: str = parameter.name

        parameter_value_name: str = param_value.__class__.__name__
        if param_value is parameter.empty:
            param_str: str = f"{param_name}: {annotation.__name__}"
        else:
            param_str = f"{param_name}: {annotation.__name__} = {parameter_value_name}"
            if isinstance(param_value, BaseField):
                param_str += f"(alias={param_value.alias}, default={param_value.default})"
    else:
        param_str = ""

    file: Optional[str] = None
    if isinstance(_object, FuncSig):
        title: str = "def"
        if inspect.iscoroutinefunction(_object.func):
            title = "async def"
        file = inspect.getfile(_object.func)
        line: int = inspect.getsourcelines(_object.func)[1]
        error_object_name: str = _object.func.__name__
        logging.debug(
            f"""
{title} {error_object_name}(
    ...
    {param_str} <-- error
    ...
):
    pass
"""
        )
    else:
        module: Optional[ModuleType] = inspect.getmodule(_object)
        if module:
            file = module.__file__
        line = inspect.getsourcelines(_object.__class__)[1]
        error_object_name = _object.__class__.__name__
        logging.debug(f"class: `{error_object_name}`  attributes error\n    {param_str}")
    raise PaitBaseException(
        f'File "{file}",' f" line {line}," f" in {error_object_name}." f" error:{str(exception)}"
    ) from exception


def parameter_2_basemodel(parameter_value_dict: Dict["inspect.Parameter", Any]) -> BaseModel:
    """Convert all parameters into pydantic mods"""
    annotation_dict: Dict[str, Tuple[Type, Any]] = {}
    param_value_dict: Dict[str, Any] = {}
    for parameter, value in parameter_value_dict.items():
        annotation_dict[parameter.name] = (parameter.annotation, ...)
        param_value_dict[parameter.name] = value

    dynamic_model: Type[BaseModel] = create_pydantic_model(annotation_dict)
    return dynamic_model(**param_value_dict)


def get_request_value_from_parameter(
    parameter: inspect.Parameter, app_helper: "BaseAppHelper"
) -> Union[Any, Coroutine]:
    field_name: str = parameter.default.__class__.__name__.lower()
    # Note: not use hasattr with LazyProperty (
    #   because hasattr will calling getattr(obj, name) and catching AttributeError,
    # )
    app_field_func: Optional[Callable] = getattr(app_helper, field_name, None)
    if app_field_func is None:
        raise NotFoundFieldError(f"field: {field_name} not found in {app_helper}")
    return app_field_func()


def _set_parameter_value_to_args(parameter: inspect.Parameter, app_helper: "BaseAppHelper", func_args: list) -> bool:
    """use func_args param faster return and extend func_args"""
    if parameter.name == "self" and not func_args:
        # Only support self param name
        func_args.append(app_helper.cbv_class)
    elif app_helper.check_request_type(parameter.annotation):
        # support request param(def handle(request: Request))
        func_args.append(app_helper.request)
    elif issubclass(parameter.annotation, PaitBaseModel):
        return True
    else:
        logging.warning(f"Pait not support args: {parameter}")
    return False


def set_parameter_value_to_args(parameter: inspect.Parameter, app_helper: "BaseAppHelper", func_args: list) -> None:
    """use func_args param faster return and extend func_args"""
    if _set_parameter_value_to_args(parameter, app_helper, func_args):
        # support pait_model param(def handle(model: PaitBaseModel))
        single_field_dict: Dict["inspect.Parameter", Any] = {}
        _pait_model: Type[PaitBaseModel] = parameter.annotation
        for param_name, param_annotation in get_type_hints(_pait_model).items():
            if param_name.startswith("_"):
                continue
            parameter = inspect.Parameter(
                param_name,
                inspect.Parameter.POSITIONAL_ONLY,
                default=getattr(_pait_model, param_name),
                annotation=param_annotation,
            )
            request_value: Any = get_request_value_from_parameter(parameter, app_helper)
            # PaitModel's attributes's annotation support BaseModel
            request_value_handle(parameter, request_value, None, single_field_dict, app_helper)
        func_args.append(parameter_2_basemodel(single_field_dict))


async def async_set_value_to_args(parameter: inspect.Parameter, app_helper: "BaseAppHelper", func_args: list) -> None:
    """use func_args param faster return and extend func_args"""
    if _set_parameter_value_to_args(parameter, app_helper, func_args):
        # support pait_model param(def handle(model: PaitModel))
        single_field_dict: Dict["inspect.Parameter", Any] = {}
        _pait_model: Type[PaitBaseModel] = parameter.annotation

        for param_name, param_annotation in get_type_hints(_pait_model).items():
            if param_name.startswith("_"):
                continue
            parameter = inspect.Parameter(
                param_name,
                inspect.Parameter.POSITIONAL_ONLY,
                default=getattr(_pait_model, param_name),
                annotation=param_annotation,
            )
            request_value: Any = get_request_value_from_parameter(parameter, app_helper)
            if asyncio.iscoroutine(request_value):
                request_value = await request_value
            # PaitModel's attributes  support BaseModel
            request_value_handle(parameter, request_value, None, single_field_dict, app_helper)
        func_args.append(parameter_2_basemodel(single_field_dict))


def request_value_handle(
    parameter: inspect.Parameter,
    request_value: Any,
    base_model_dict: Optional[Dict[str, Any]],
    parameter_value_dict: Dict["inspect.Parameter", Any],
    app_helper: "BaseAppHelper",
) -> None:
    """parse request_value and set to base_model_dict or parameter_value_dict"""
    param_value: BaseField = parameter.default
    annotation: Type[BaseModel] = parameter.annotation
    param_name: str = parameter.name

    if not isinstance(param_value, BaseField):
        raise PaitBaseException(f"must use {BaseField.__class__.__name__}, no {param_value}")
    if param_value.raw_return:
        parameter_value_dict[parameter] = request_value
    elif (
        isinstance(request_value, Mapping)
        or app_helper.check_header_type(type(request_value))
        or app_helper.check_form_type(type(request_value))
    ):

        if base_model_dict is not None and inspect.isclass(annotation) and issubclass(annotation, BaseModel):
            # parse annotation is pydantic.BaseModel and base_model_dict not None
            base_model_dict[parameter.name] = annotation(**request_value)
        else:
            # parse annotation is python type and pydantic.field
            if type(param_value.alias) is str and param_value.alias in request_value:
                value = request_value[param_value.alias]
            elif param_name in request_value:
                value = request_value[param_name]
            else:
                # not found value
                if param_value.default != fields.Undefined:
                    value = param_value.default
                else:
                    parameter_value_name: str = param_value.__class__.__name__
                    param_str: str = (
                        f"{param_name}: {annotation} = {parameter_value_name}("
                        f"alias={param_value.alias}, default={param_value.default})"
                    )
                    raise NotFoundValueError(
                        f"kwargs param:{param_str} not found in {request_value},"
                        f" try use {parameter_value_name}(alias={{alias}})"
                    )
            parameter_value_dict[parameter] = value
    else:
        parameter_value_dict[parameter] = request_value


def param_handle(
    app_helper: "BaseAppHelper", _object: Union[FuncSig, Type], param_list: List["inspect.Parameter"]
) -> Tuple[List[Any], Dict[str, Any]]:
    args_param_list: List[Any] = []
    kwargs_param_dict: Dict[str, Any] = {}
    single_field_dict: Dict["inspect.Parameter", Any] = {}

    for parameter in param_list:
        try:
            if parameter.default != parameter.empty:
                # kwargs param
                # support model: def demo(pydantic.BaseModel: BaseModel = pait.field.BaseField())
                if isinstance(parameter.default, field.Depends):
                    func: Callable = parameter.default.func
                    func_sig: FuncSig = get_func_sig(func)
                    _func_args, _func_kwargs = param_handle(app_helper, func_sig, func_sig.param_list)
                    func_result: Any = func(*_func_args, **_func_kwargs)
                    kwargs_param_dict[parameter.name] = func_result
                else:
                    request_value: Any = get_request_value_from_parameter(parameter, app_helper)
                    request_value_handle(parameter, request_value, kwargs_param_dict, single_field_dict, app_helper)
            else:
                # args param
                # support model: model: ModelType
                set_parameter_value_to_args(parameter, app_helper, args_param_list)
        except PaitBaseException as e:
            raise_and_tip(_object, e, parameter)
    # support field: def demo(demo_param: int = pait.field.BaseField())
    if single_field_dict:
        try:
            kwargs_param_dict.update(parameter_2_basemodel(single_field_dict).dict())
        except Exception as e:
            raise_and_tip(_object, e)
    return args_param_list, kwargs_param_dict


async def async_param_handle(
    dispatch_web: "BaseAppHelper", _object: Union[FuncSig, Type], param_list: List["inspect.Parameter"]
) -> Tuple[List[Any], Dict[str, Any]]:
    args_param_list: List[Any] = []
    kwargs_param_dict: Dict[str, Any] = {}
    single_field_dict: Dict["inspect.Parameter", Any] = {}
    for parameter in param_list:
        try:
            if parameter.default != parameter.empty:
                # kwargs param
                # support model: pydantic.BaseModel = pait.field.BaseField()
                if isinstance(parameter.default, field.Depends):
                    func: Callable = parameter.default.func
                    func_sig: FuncSig = get_func_sig(func)
                    _func_args, _func_kwargs = await async_param_handle(dispatch_web, func_sig, func_sig.param_list)
                    func_result: Any = func(*_func_args, **_func_kwargs)
                    if asyncio.iscoroutine(func_result):
                        func_result = await func_result
                    kwargs_param_dict[parameter.name] = func_result
                else:
                    request_value: Any = get_request_value_from_parameter(parameter, dispatch_web)

                    if asyncio.iscoroutine(request_value) or asyncio.isfuture(request_value):
                        request_value = await request_value

                    request_value_handle(parameter, request_value, kwargs_param_dict, single_field_dict, dispatch_web)
            else:
                # args param
                # support model: model: ModelType
                await async_set_value_to_args(parameter, dispatch_web, args_param_list)
        except PaitBaseException as e:
            raise_and_tip(_object, e, parameter)

    # support field: def demo(demo_param: int = pait.field.BaseField())
    if single_field_dict:
        try:
            try:
                kwargs_param_dict.update(parameter_2_basemodel(single_field_dict).dict())
            except TypeError:
                kwargs_param_dict.update(parameter_2_basemodel(single_field_dict).__dict__)

        except Exception as e:
            raise_and_tip(_object, e)
    return args_param_list, kwargs_param_dict


async def async_class_param_handle(dispatch_web: "BaseAppHelper") -> None:
    cbv_class: Optional[Type] = dispatch_web.cbv_class
    if not cbv_class:
        return
    param_list: list = get_parameter_list_from_class(cbv_class)
    args, kwargs = await async_param_handle(dispatch_web, cbv_class, param_list)
    cbv_class.__dict__.update(kwargs)


def class_param_handle(dispatch_web: "BaseAppHelper") -> None:
    cbv_class: Optional[Type] = dispatch_web.cbv_class
    if not cbv_class:
        return
    param_list: list = get_parameter_list_from_class(cbv_class)
    args, kwargs = param_handle(dispatch_web, cbv_class, param_list)
    cbv_class.__dict__.update(kwargs)


async def async_func_param_handle(dispatch_app: "BaseAppHelper", func_sig: FuncSig) -> Tuple[List[Any], Dict[str, Any]]:
    return await async_param_handle(dispatch_app, func_sig, func_sig.param_list)


def func_param_handle(dispatch_app: "BaseAppHelper", func_sig: FuncSig) -> Tuple[List[Any], Dict[str, Any]]:
    return param_handle(dispatch_app, func_sig, func_sig.param_list)



from typing import Any, Dict, Optional

import yaml

from pait.model import PaitCoreModel

from .base_parse import PaitBaseParse  # type: ignore


class PaitYaml(PaitBaseParse):
    def __init__(self, pait_dict: Dict[str, PaitCoreModel], title: str = "Pait Yaml", filename: Optional[str] = None):
        super().__init__(pait_dict, undefined="Required")

        _pait_dict: Dict[str, Any] = self.gen_dict()
        _pait_dict["title"] = title

        pait_yaml: str = yaml.dump(_pait_dict)
        self.output(filename, pait_yaml, ".yaml")



import json
from typing import Any, Dict, Optional

from pait.model import PaitCoreModel

from .base_parse import PaitBaseParse  # type: ignore


class PaitJson(PaitBaseParse):
    def __init__(
        self,
        pait_dict: Dict[str, PaitCoreModel],
        title: str = "Pait Json",
        filename: Optional[str] = None,
        indent: Optional[int] = None,
    ):
        super().__init__(pait_dict, undefined="Required")

        _pait_dict: Dict[str, Any] = self.gen_dict()
        _pait_dict["title"] = title

        pait_json: str = json.dumps(_pait_dict, indent=indent)
        self.output(filename, pait_json, ".json")



import inspect
import sys
from concurrent import futures
from dataclasses import dataclass
from typing import Any, Callable, Dict, ForwardRef, List, Optional, Tuple, Type, get_type_hints

from pydantic import BaseModel, create_model


class UndefinedType:
    def __repr__(self) -> str:
        return "PaitUndefined"


# pait undefined flag
Undefined: UndefinedType = UndefinedType()


class LazyProperty:
    """Cache field computing resources
    >>> class Demo:
    ...     @LazyProperty
    ...     def value(self, value):
    ...         return value * value
    """

    def __call__(self, func: Callable) -> Callable:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            class_ = args[0]
            future: Optional[futures.Future] = getattr(
                class_, f"{self.__class__.__name__}_{func.__name__}_future", None
            )
            if not future:
                future = futures.Future()
                result: Any = func(*args, **kwargs)
                future.set_result(result)
                setattr(class_, f"{self.__class__.__name__}_{func.__name__}_future", future)
                return result
            return future.result()

        return wrapper


def create_pydantic_model(annotation_dict: Dict[str, Tuple[Type, Any]]) -> Type[BaseModel]:
    """pydantic create_model helper
    if use create_model('DynamicModel', **annotation_dict), mypy will tip error
    """
    return create_model(
        "DynamicModel",
        __config__=None,
        __base__=None,
        __module__="pydantic.main",
        __validators__=None,
        **annotation_dict,
    )


@dataclass()
class FuncSig:
    """func inspect.Signature model"""

    func: Callable
    sig: "inspect.Signature"
    param_list: List["inspect.Parameter"]


def get_func_sig(func: Callable) -> FuncSig:
    """get func inspect.Signature model"""
    sig: inspect.Signature = inspect.signature(func)
    param_list: List[inspect.Parameter] = []
    for key in sig.parameters:
        if not (sig.parameters[key].annotation != sig.empty or sig.parameters[key].name == "self"):
            continue
        parameter: inspect.Parameter = sig.parameters[key]
        if isinstance(parameter.annotation, str):
            value: ForwardRef = ForwardRef(parameter.annotation, is_argument=False)
            setattr(
                parameter, "_annotation", value._evaluate(sys.modules[func.__module__].__dict__, None)  # type: ignore
            )
        param_list.append(parameter)

    # return_param = sig.return_annotation
    return FuncSig(func=func, sig=sig, param_list=param_list)


def get_parameter_list_from_class(cbv_class: Type) -> List["inspect.Parameter"]:
    """get class parameter list by attributes, if attributes not default value, it will be set `Undefined`"""
    parameter_list: List["inspect.Parameter"] = []
    if hasattr(cbv_class, "__annotations__"):
        for param_name, param_annotation in get_type_hints(cbv_class).items():
            parameter: "inspect.Parameter" = inspect.Parameter(
                param_name,
                inspect.Parameter.POSITIONAL_ONLY,
                default=getattr(cbv_class, param_name, Undefined),
                annotation=param_annotation,
            )
            parameter_list.append(parameter)
    return parameter_list



import inspect
import warnings
from types import CodeType
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, get_type_hints

from pydantic import BaseModel
from pydantic.fields import Undefined

from pait.field import BaseField, Depends
from pait.model import PaitBaseModel, PaitCoreModel, PaitResponseModel
from pait.util import FuncSig, create_pydantic_model, get_func_sig, get_parameter_list_from_class


class PaitBaseParse(object):
    def __init__(self, pait_dict: Dict[str, PaitCoreModel], undefined: Any = Undefined):
        self._undefined: Any = undefined
        self._group_list: List[str] = []
        self._group_pait_dict: Dict[str, List[PaitCoreModel]] = {}

        self._init(pait_dict)

    def _init(self, pait_dict: Dict[str, PaitCoreModel]) -> None:
        """read from `pait_id_dict` and write PaitMd attributes"""
        for pait_id, pait_model in pait_dict.items():
            if not pait_model.operation_id:
                continue
            group: str = pait_model.group
            if group not in self._group_pait_dict:
                self._group_pait_dict[group] = [pait_model]
            else:
                self._group_pait_dict[group].append(pait_model)
        self._group_list = sorted(self._group_pait_dict.keys())

    def _parse_schema(
        self, schema_dict: dict, definition_dict: Optional[dict] = None, parent_key: str = ""
    ) -> List[dict]:
        """gen pait field dict from pydantic basemodel schema"""
        field_dict_list: List[dict] = []
        # model property openapi dict
        # e.g. : {'code': {'title': 'Code', 'description': 'api code', 'default': 1, 'type': 'integer'}}
        property_dict: dict = schema_dict["properties"]
        # class schema in the parent schema
        if not definition_dict:
            definition_dict = schema_dict.get("definitions", {})
        for param_name, param_dict in property_dict.items():
            if parent_key:
                all_param_name: str = f"{parent_key}.{param_name}"
            else:
                all_param_name = param_name

            if "$ref" in param_dict and definition_dict:
                # ref support
                key: str = param_dict["$ref"].split("/")[-1]
                if isinstance(definition_dict, dict):
                    field_dict_list.extend(self._parse_schema(definition_dict[key], definition_dict, all_param_name))
            elif "items" in param_dict and "$ref" in param_dict["items"]:
                # mad item ref support
                key = param_dict["items"]["$ref"].split("/")[-1]
                if isinstance(definition_dict, dict):
                    field_dict_list.extend(self._parse_schema(definition_dict[key], definition_dict, all_param_name))
            else:
                if "enum" in param_dict:
                    # enum support
                    default: str = param_dict.get("enum", self._undefined)
                    if default is not self._undefined:
                        default = f'Only choose from: {",".join(["`" + i + "`" for i in default])}'
                    _type: str = "enum"
                else:
                    default = param_dict.get("default", self._undefined)
                    _type = param_dict["type"]
                field_dict_list.append(
                    {
                        "param_name": all_param_name,
                        "description": param_dict.get("description", ""),
                        "default": default,
                        "type": _type,
                        "other": {
                            key: value
                            for key, value in param_dict.items()
                            if key not in {"description", "title", "type", "default"}
                        },
                        "raw": {
                            "param_name": param_name,
                            "schema": param_dict,
                            "parent_schema": schema_dict,
                        },
                    }
                )
        return field_dict_list

    def _parse_base_model_to_field_dict(
        self,
        field_dict: Dict[str, List[Dict[str, Any]]],
        _pydantic_model: Type[BaseModel],
        _pait_field_dict: Dict[str, BaseField],
    ) -> None:
        """
        write field dict from _pydantic_model or _pait_field_dict
        :param field_dict:
          e.g.
            {
                [
                    {
                        "param_name": "",
                        "description": "",
                        "default": "",
                        "type": _"",
                        "other": {"": ""},
                        "raw": {
                            "param_name": "",
                            "schema": {"": ""},
                            "parent_schema": {"": ""},
                        },
                    }
                ]
            }
        :param _pydantic_model: pydantic.basemodel
        :param _pait_field_dict:
            e.g.
            {
                'uid': Query(default=Ellipsis, description='user id', gt=10, lt=1000, extra={}),
                'user_name': Query(default=Ellipsis, description='user name', min_length=2, max_length=4, extra={}),
                'user_agent': Header(default=Ellipsis, alias='user-agent', alias_priority=2, description='user agent', extra={}),
                'age': Body(default=Ellipsis, description='age', gt=1, lt=100, extra={})
            }
        :return:
        """
        # TODO design like _parse_schema
        param_name_alias_dict: Dict[str, str] = {
            value.alias: key for key, value in _pait_field_dict.items() if isinstance(value, BaseField) and value.alias
        }
        property_dict: Dict[str, Any] = _pydantic_model.schema()["properties"]
        for param_name, param_dict in property_dict.items():
            param_python_name: str = param_name_alias_dict.get(param_name, param_name)
            field_name = _pait_field_dict[param_python_name].__class__.__name__.lower()
            if "$ref" in param_dict:
                # ref support
                key: str = param_dict["$ref"].split("/")[-1]
                param_dict = _pydantic_model.schema()["definitions"][key]
            elif "items" in param_dict and "$ref" in param_dict["items"]:
                # mad item ref support
                key = param_dict["items"]["$ref"].split("/")[-1]
                param_dict = _pydantic_model.schema()["definitions"][key]
            elif "allOf" in param_dict:
                if len(param_dict["allOf"]) > 1:
                    warnings.warn(f"{param_dict['param_name']} only support 1 item")

                param_dict.update(param_dict["allOf"][0])
                key = param_dict["$ref"].split("/")[-1]
                param_dict = _pydantic_model.schema()["definitions"][key]
            if "enum" in param_dict:
                # enum support
                default: str = param_dict.get("enum", self._undefined)
                if default is not self._undefined:
                    default = f'Only choose from: {",".join(["`" + i + "`" for i in default])}'
                _type: str = "enum"
                description: str = _pait_field_dict[param_python_name].description
            else:
                default = param_dict.get("default", self._undefined)
                _type = param_dict.get("type", self._undefined)
                description = param_dict.get("description")
            _field_dict = {
                "param_name": param_name,
                "description": description,
                "default": default,
                "type": _type,
                "other": {
                    key: value
                    for key, value in param_dict.items()
                    if key not in {"description", "title", "type", "default"}
                },
                "raw": {
                    "param_name": param_name,
                    "schema": param_dict,
                    "parent_schema": _pydantic_model.schema(),
                    "annotation": _pydantic_model.__annotations__[param_python_name],
                    "field": _pait_field_dict[param_python_name],
                },
            }
            if field_name not in field_dict:
                field_dict[field_name] = [_field_dict]
            else:
                field_dict[field_name].append(_field_dict)

    def parameter_list_handle(
        self,
        parameter_list: List["inspect.Parameter"],
        field_dict: Dict[str, List[Dict[str, Any]]],
        single_field_list: List[Tuple[str, "inspect.Parameter"]],
    ) -> None:
        """write field_dict or single_field_list from parameter_list"""
        for parameter in parameter_list:
            if parameter.default != parameter.empty:
                annotation: type = parameter.annotation
                if inspect.isclass(annotation) and issubclass(annotation, BaseModel):
                    # def test(test_model: BaseModel = Body())
                    _pait_field_dict: Dict[str, BaseField] = {
                        param_name: parameter.default
                        for param_name, annotation in get_type_hints(annotation).items()
                        if not param_name.startswith("_")
                    }
                    self._parse_base_model_to_field_dict(field_dict, annotation, _pait_field_dict)
                else:
                    # def test(test_model: int = Body())
                    if isinstance(parameter.default, Depends):
                        field_dict.update(self._parse_func_param_to_field_dict(parameter.default.func))
                    else:
                        field_name: str = parameter.default.__class__.__name__.lower()
                        single_field_list.append((field_name, parameter))
            elif issubclass(parameter.annotation, PaitBaseModel):
                # def test(test_model: PaitBaseModel)
                _pait_model: Type[PaitBaseModel] = parameter.annotation
                _pait_field_dict = {
                    param_name: getattr(_pait_model, param_name)
                    for param_name, param_annotation in get_type_hints(_pait_model).items()
                    if not param_name.startswith("_")
                }
                pait_base_model = _pait_model.to_pydantic_model()
                self._parse_base_model_to_field_dict(field_dict, pait_base_model, _pait_field_dict)

    def _parse_func_param_to_field_dict(self, func: Callable) -> Dict[str, List[Dict[str, Any]]]:
        """gen filed dict from func
        [
            {
                'field': {
                    'param_name': str,
                    'description': str,
                    'default': str,
                    'type': type,
                    'other': dict,
                    'raw': {
                        'param_name': str,
                        'schema': dict,
                        'parent_schema': pydantic base model.schema(),
                        'annotation': annotation,
                        'field': basefield,
                    }
                }
            }
        ]
        """
        field_dict: Dict[str, List[Dict[str, Any]]] = {}
        func_sig: FuncSig = get_func_sig(func)
        single_field_list: List[Tuple[str, "inspect.Parameter"]] = []

        qualname = func.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0]
        class_ = getattr(inspect.getmodule(func), qualname)
        if inspect.isclass(class_):
            parameter_list: List["inspect.Parameter"] = get_parameter_list_from_class(class_)
            self.parameter_list_handle(parameter_list, field_dict, single_field_list)
        self.parameter_list_handle(func_sig.param_list, field_dict, single_field_list)

        if single_field_list:
            annotation_dict: Dict[str, Tuple[Type, Any]] = {}
            _pait_field_dict: Dict[str, BaseField] = {}
            for field_name, parameter in single_field_list:
                field: BaseField = parameter.default
                annotation_dict[parameter.name] = (parameter.annotation, field)
                _pait_field_dict[parameter.name] = field

            _pydantic_model: Type[BaseModel] = create_pydantic_model(annotation_dict)
            self._parse_base_model_to_field_dict(field_dict, _pydantic_model, _pait_field_dict)
        return field_dict

    def gen_dict(self) -> Dict[str, Any]:
        """gen pait api dict"""
        api_doc_dict: Dict[str, Any] = {"group": {}}
        for group in self._group_list:
            group_dict: Dict[str, Any] = {"name": group, "group_list": []}
            api_doc_dict["group"][group] = group_dict
            for pait_model in self._group_pait_dict[group]:
                func_code: CodeType = pait_model.func.__code__
                response_list: List[Dict[str, Any]] = []
                if pait_model.response_model_list:
                    for resp_model_class in pait_model.response_model_list:
                        resp_model: PaitResponseModel = resp_model_class()
                        if resp_model.response_data:
                            schema_dict: dict = resp_model.response_data.schema()
                        else:
                            schema_dict = {}
                        response_list.append(
                            {
                                "status_code": ",".join([str(i) for i in resp_model.status_code]),
                                "media_type": resp_model.media_type,
                                "description": resp_model.description,
                                "header": resp_model.header,
                                "response_data": self._parse_schema(schema_dict),
                            }
                        )
                field_dict: Dict[str, List[Dict[str, Any]]] = self._parse_func_param_to_field_dict(pait_model.func)
                for field_name, field_dict_list in field_dict.items():
                    for field_dict in field_dict_list:
                        del field_dict["raw"]

                group_dict["group_list"].append(
                    {
                        "name": pait_model.operation_id,
                        "status": pait_model.status.value,
                        "author": ",".join(pait_model.author),
                        "func": {
                            "file": func_code.co_filename,
                            "lineno": func_code.co_firstlineno,
                            "name": pait_model.func.__qualname__,
                        },
                        "path": pait_model.path,
                        "method": pait_model.method_list,
                        "request": field_dict,
                        "response_list": response_list,
                    }
                )
        return api_doc_dict

    @staticmethod
    def output(filename: Optional[str], content: str, suffix: str) -> None:
        if not filename:
            print(content)
        else:
            if not filename.endswith(suffix):
                filename += suffix
            with open(filename, mode="w") as f:
                f.write(content)



import logging
from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Type

from sanic.app import Sanic
from sanic.headers import HeaderIterable
from sanic.request import File, Request, RequestParameters

from pait.app.base import BaseAppHelper
from pait.core import pait as _pait
from pait.g import pait_data
from pait.model import PaitCoreModel, PaitResponseModel, PaitStatus
from pait.util import LazyProperty


class AppHelper(BaseAppHelper):
    RequestType = Request
    FormType = RequestParameters
    FileType = File
    HeaderType = HeaderIterable
    app_name = "sanic"

    def __init__(self, class_: Any, args: Tuple[Any, ...], kwargs: Mapping[str, Any]):
        super().__init__(class_, args, kwargs)

    def body(self) -> dict:
        return self.request.json

    def cookie(self) -> dict:
        return self.request.cookies

    def file(self) -> Dict[str, File]:
        return {key: value[0] for key, value in self.request.files.items()}

    def form(self) -> dict:
        return {key: value[0] for key, value in self.request.form.items()}

    def header(self) -> HeaderIterable:
        return self.request.headers

    def path(self) -> Mapping[str, Any]:
        return self.request_kwargs

    @LazyProperty()
    def query(self) -> dict:
        return {key: value[0] for key, value in self.request.args.items()}

    @LazyProperty()
    def multiform(self) -> Dict[str, List[Any]]:
        return {key: self.request.form.getlist(key) for key, _ in self.request.form.items()}

    @LazyProperty()
    def multiquery(self) -> Dict[str, Any]:
        return {key: self.request.args.getlist(key) for key, _ in self.request.args.items()}


def load_app(app: Sanic) -> Dict[str, PaitCoreModel]:
    _pait_data: Dict[str, PaitCoreModel] = {}
    """Read data from the route that has been registered to `pait`"""
    for parts, route in app.router.routes_all.items():
        if "static" in route.name:
            continue

        route_name: str = route.name
        for path, handler_dict in route.handlers.items():
            for method, handler_list in handler_dict.items():
                for handler in handler_list:
                    view_class: Optional[Type] = getattr(handler, "view_class", None)
                    if view_class:
                        handler = getattr(view_class, method.lower(), None)
                    if not handler:
                        logging.warning(f"{route_name} can not found handle")
                        continue
                    pait_id: Optional[str] = getattr(handler, "_pait_id", None)
                    if not pait_id:
                        logging.warning(f"{route_name} can not found pait id")
                        continue
                    pait_data.add_route_info(AppHelper.app_name, pait_id, path, {method}, route_name)
                    _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    return _pait_data


def pait(
    author: Optional[Tuple[str]] = None,
    desc: Optional[str] = None,
    status: Optional[PaitStatus] = None,
    group: Optional[str] = None,
    tag: Optional[Tuple[str, ...]] = None,
    response_model_list: List[Type[PaitResponseModel]] = None,
) -> Callable:
    """Help starlette provide parameter checks and type conversions for each routing function/cbv class"""
    return _pait(
        AppHelper,
        author=author,
        desc=desc,
        status=status,
        group=group,
        tag=tag,
        response_model_list=response_model_list,
    )



from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from pait.model import PaitCoreModel, PaitResponseModel, PaitStatus

from .auto_load_app import auto_load_app_class  # type: ignore


def load_app(app: Any) -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`
    Note:This is an implicit method
    """
    app_name: str = app.__class__.__name__.lower()
    if app_name == "flask":
        from .flask import load_app  # type: ignore

        return load_app(app)
    elif app_name == "starlette":
        from .starlette import load_app  # type: ignore

        return load_app(app)
    elif app_name == "sanic":
        from .sanic import load_app  # type: ignore

        return load_app(app)
    elif app_name == "application" and app.__class__.__module__ == "tornado.web":
        from .tornado import load_app  # type: ignore

        return load_app(app)
    else:
        raise NotImplementedError(f"Pait not support:{app}")


def pait(
    author: Optional[Tuple[str]] = None,
    desc: Optional[str] = None,
    status: Optional[PaitStatus] = None,
    group: str = "root",
    tag: Optional[Tuple[str, ...]] = None,
    response_model_list: List[Type[PaitResponseModel]] = None,
) -> Callable:
    """provide parameter checks and type conversions for each routing function/cbv class
    Note:This is an implicit method
    """
    load_class_app = auto_load_app_class()
    app_name: str = load_class_app.__name__.lower()
    if app_name == "flask":
        from .flask import pait
    elif app_name == "starlette":
        from .starlette import pait
    elif app_name == "sanic":
        from .sanic import pait
    elif app_name == "tornado":
        from .tornado import pait
    else:
        raise NotImplementedError(f"Pait not support:{load_class_app}")
    return pait(author, desc, status, group, tag, response_model_list)



from typing import Any, Callable, Coroutine, Dict, List, Mapping, Optional, Tuple, Type, Union

from starlette.applications import Starlette
from starlette.datastructures import FormData, Headers, UploadFile
from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.routing import Route

from pait.app.base import BaseAppHelper
from pait.core import pait as _pait
from pait.g import pait_data
from pait.model import PaitCoreModel, PaitResponseModel, PaitStatus
from pait.util import LazyProperty


class AppHelper(BaseAppHelper):
    RequestType = Request
    FormType = FormData
    FileType = UploadFile
    HeaderType = Headers
    app_name = "starlette"

    def __init__(self, class_: Any, args: Tuple[Any, ...], kwargs: Mapping[str, Any]):
        super().__init__(class_, args, kwargs)

    def body(self) -> dict:
        return self.request.json()

    def cookie(self) -> dict:
        return self.request.cookies

    def file(self) -> UploadFile:
        return self.request.form()

    def form(self) -> FormData:
        return self.request.form()

    def header(self) -> Headers:
        return self.request.headers

    def path(self) -> dict:
        return self.request.path_params

    def query(self) -> dict:
        return dict(self.request.query_params)

    @LazyProperty()
    def multiform(self) -> Coroutine[Any, Any, Dict[str, List[Any]]]:
        async def _() -> Dict[str, List[Any]]:
            form_data = await self.request.form()
            return {key: form_data.getlist(key) for key, _ in form_data.items()}

        return _()

    @LazyProperty()
    def multiquery(self) -> Dict[str, Any]:
        return {key: self.request.query_params.getlist(key) for key, _ in self.request.query_params.items()}


def load_app(app: Starlette) -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`"""
    _pait_data: Dict[str, PaitCoreModel] = {}
    for route in app.routes:
        if not isinstance(route, Route):
            # not support
            continue
        path: str = route.path
        method_set: set = route.methods or set()
        route_name: str = route.name
        endpoint: Union[Callable, Type] = route.endpoint
        pait_id: str = getattr(route.endpoint, "_pait_id", None)
        if not pait_id and issubclass(endpoint, HTTPEndpoint):  # type: ignore
            for method in ["get", "post", "head", "options", "delete", "put", "trace", "patch"]:
                method_endpoint = getattr(endpoint, method, None)
                if not method_endpoint:
                    continue
                method_set = {method}
                pait_id = getattr(method_endpoint, "_pait_id", None)
                if not pait_id:
                    continue
                pait_data.add_route_info(AppHelper.app_name, pait_id, path, method_set, f"{route_name}.{method}")
                _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
        else:
            pait_data.add_route_info(AppHelper.app_name, pait_id, path, method_set, route_name)
            _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    return _pait_data


def pait(
    author: Optional[Tuple[str]] = None,
    desc: Optional[str] = None,
    status: Optional[PaitStatus] = None,
    group: Optional[str] = None,
    tag: Optional[Tuple[str, ...]] = None,
    response_model_list: List[Type[PaitResponseModel]] = None,
) -> Callable:
    """Help starlette provide parameter checks and type conversions for each routing function/cbv class"""
    return _pait(
        AppHelper,
        author=author,
        desc=desc,
        status=status,
        group=group,
        tag=tag,
        response_model_list=response_model_list,
    )



import json
from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Type

from tornado.httputil import RequestStartLine
from tornado.web import Application

from pait.app.base import BaseAppHelper
from pait.core import pait as _pait
from pait.g import pait_data
from pait.model import PaitCoreModel, PaitResponseModel, PaitStatus
from pait.util import LazyProperty


class AppHelper(BaseAppHelper):
    RequestType = RequestStartLine
    FormType = dict
    FileType = dict
    HeaderType = dict
    app_name = "tornado"

    def __init__(self, class_: Any, args: Tuple[Any, ...], kwargs: Mapping[str, Any]):
        super().__init__(class_, args, kwargs)
        if not self.cbv_class:
            raise RuntimeError("Can not load Tornado handle")

        self.request = self.cbv_class.request
        self.path_kwargs: Dict[str, Any] = self.cbv_class.path_kwargs

    def body(self) -> dict:
        return json.loads(self.request.body.decode())

    def cookie(self) -> dict:
        return self.request.cookies

    def file(self) -> dict:
        return self.request.files

    def form(self) -> dict:
        return {key: value[0] for key, value in self.request.body_arguments}

    def header(self) -> dict:
        return self.request.headers

    def path(self) -> dict:
        return self.path_kwargs

    @LazyProperty()
    def query(self) -> dict:
        return {key: value[0].decode() for key, value in self.request.query_arguments.items()}

    @LazyProperty()
    def multiform(self) -> Dict[str, List[Any]]:
        return self.request.body_arguments

    @LazyProperty()
    def multiquery(self) -> Dict[str, Any]:
        return {key: [i.decode() for i in value] for key, value in self.request.query_arguments.items()}


def load_app(app: Application) -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`"""
    _pait_data: Dict[str, PaitCoreModel] = {}
    for rule in app.wildcard_router.rules:
        path: str = rule.matcher.regex.pattern  # type: ignore
        base_name: str = rule.target.__name__
        for method in ["get", "post", "head", "options", "delete", "put", "trace", "patch"]:
            try:
                handler: Callable = getattr(rule.target, method, None)
            except TypeError:
                continue
            route_name: str = f"{base_name}.{method}"
            pait_id: Optional[str] = getattr(handler, "_pait_id", None)
            if not pait_id:
                continue
            pait_data.add_route_info(AppHelper.app_name, pait_id, path, {method}, route_name)
            _pait_data[pait_id] = pait_data.get_pait_data(AppHelper.app_name, pait_id)
    return _pait_data


def pait(
    author: Optional[Tuple[str]] = None,
    desc: Optional[str] = None,
    status: Optional[PaitStatus] = None,
    group: str = "root",
    tag: Optional[Tuple[str, ...]] = None,
    response_model_list: List[Type[PaitResponseModel]] = None,
) -> Callable:
    """Help starlette provide parameter checks and type conversions for each routing function/cbv class"""
    return _pait(
        AppHelper,
        author=author,
        desc=desc,
        status=status,
        group=group,
        tag=tag,
        response_model_list=response_model_list,
    )



import logging
from typing import Any, Callable, Dict, Set

from pait.model import PaitCoreModel


class PaitData(object):
    def __init__(self) -> None:
        self.pait_id_dict: Dict[str, Dict[str, "PaitCoreModel"]] = {}

    def register(self, app_name: str, pait_info_model: "PaitCoreModel") -> None:
        pait_id: str = pait_info_model.pait_id
        if app_name not in self.pait_id_dict:
            self.pait_id_dict[app_name] = {}
        self.pait_id_dict[app_name][pait_id] = pait_info_model

    def get_pait_data(self, app_name: str, pait_id: str) -> PaitCoreModel:
        return self.pait_id_dict[app_name][pait_id]

    def add_route_info(self, app_name: str, pait_id: str, path: str, method_set: Set[str], route_name: str) -> None:
        if pait_id in self.pait_id_dict[app_name]:
            self.pait_id_dict[app_name][pait_id].path = path
            self.pait_id_dict[app_name][pait_id].method_list = sorted(list(method_set or set()))
            self.pait_id_dict[app_name][pait_id].operation_id = route_name
        else:
            logging.warning(f"loan path:{path} fail, pait id:{pait_id}")

    #
    # def __getitem__(self, item: Any) -> Any:
    #     return self.pait_id_dict[item]
    #
    # def __setitem__(self, key: str, value: "PaitCoreModel") -> None:
    #     self.pait_id_dict[key] = value

    def __bool__(self) -> bool:
        return bool(self.pait_id_dict)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pait', 'pait.api_doc', 'pait.app']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0', 'pydantic>=1.7.3,<2.0.0']

setup_kwargs = {
    'name': 'pait',
    'version': '0.5.3',
    'description': 'Pait is a Python api tool. Pait enables your Python web framework to have type checking, parameter type conversion and interface document generation (power by inspect, pydantic)',
    'long_description': '# pait\nPait is an api tool that can be used in any python web framework (currently only `flask`, `starlette`, `sanic`, `tornado` are supported, other frameworks will be supported once Pait is stable).\n\nThe core functionality of Pait is to allow you to have FastAPI-like type checking and type conversion functionality (dependent on Pydantic and inspect) in any Python web framework, as well as documentation output\n\nPait\'s vision of documentation output is both code and documentation, with a simple configuration, you can get an md document or openapi (json, yaml)\n\n> Note:\n> \n> mypy check 100%\n> \n> test coverage 95%+ (exclude api_doc)\n> \n> python version >= 3.7 (support postponed annotations)\n \n[中文文档](https://github.com/so1n/pait/blob/master/README_ZH.md)\n# Installation\n```Bash\npip install pait\n```\n# Usage\nNote: The following code does not specify, all default to use the `starlette` framework. \nNote: There is no test case for the document output function, and the function is still being improved\n## 1.type checking and parameter type conversion\n### 1.1.Use in route handle\nA simple starlette route handler example:\n```Python\nimport uvicorn\n\nfrom starlette.applications import Starlette\nfrom starlette.routing import Route\nfrom starlette.requests import Request\nfrom starlette.responses import JSONResponse\n\n\nasync def demo_post(request: Request) -> JSONResponse:\n    body_dict: dict = await request.json()\n    uid: int = body_dict.get(\'uid\')\n    user_name: str = body_dict.get(\'user_name\')\n    # The following code is only for demonstration, in general, we do some wrapping \n    if not uid:\n        raise ValueError(\'xxx\')\n    if type(uid) != int:\n        raise TypeError(\'xxxx\')\n    if 10 <= uid <= 1000:\n        raise ValueError(\'xxx\')\n\n    if not user_name:\n        raise ValueError(\'xxx\')\n    if type(user_name) != str:\n        raise TypeError(\'xxxx\')\n    if 2 <= len(user_name) <= 4:\n        raise ValueError(\'xxx\')\n    \n    return JSONResponse(\n        {\n            \'result\': {\n                \'uid\': body_dict[\'uid\'],\n                \'user_name\': body_dict[\'user_name\']\n            } \n        }\n    )\n\n\napp = Starlette(\n    routes=[\n        Route(\'/api\', demo_post, methods=[\'POST\']),\n    ]\n)\n\n\nuvicorn.run(app)\n```\nuse pait in starletter route handler:\n\n```Python\nimport uvicorn\n\nfrom starlette.applications import Starlette\nfrom starlette.routing import Route\nfrom starlette.responses import JSONResponse\n\nfrom pait.field import Body\nfrom pait.app.starlette import pait\nfrom pydantic import (\n    BaseModel,\n    conint,\n    constr,\n)\n\n\n# Create a Model based on Pydantic.BaseModel \nclass PydanticModel(BaseModel):\n    uid: conint(gt=10, lt=1000)  # Whether the auto-check type is int, and whether it is greater than or equal to 10 and less than or equal to 1000 \n    user_name: constr(min_length=2, max_length=4)  # Whether the auto-check type is str, and whether the length is greater than or equal to 2, less than or equal to 4\n\n\n\n# Decorating functions with the pait decorator\n@pait()\nasync def demo_post(\n    # pait through the Body () to know the current need to get the value of the body from the request, and assign the value to the model, \n    # and the structure of the model is the above PydanticModel, he will be based on our definition of the field automatically get the value and conversion and judgment\n    model: PydanticModel = Body.i()\n):\n    # Get the corresponding value to return\n    return JSONResponse({\'result\': model.dict()})\n\n\napp = Starlette(\n    routes=[\n        Route(\'/api\', demo_post, methods=[\'POST\']),\n    ]\n)\n\nuvicorn.run(app)\n```\nAs you can see, you just need to add a `pait` decorator to the routing function and change the parameters of `demo_post` to `model: PydanticModel = Body()`.\npait through `Body` to know the need to get the post request body data, and according to `conint(gt=10, lt=1000)` on the data conversion and restrictions, and assigned to `PydanticModel`, the user only need to use `Pydantic` as the call `model` can get the data.\n\nHere is just a simple demo, because we write the model can be reused, so you can save a lot of development time, the above parameters are only used to a way to write, the following will introduce pait support for the two ways to write and use.\n\n### 1.2.Parameter expression supported by pait\npait in order to facilitate the use of users, support a variety of writing methods (mainly the difference between TypeHints)\n- TypeHints  is PaitBaseModel, mainly used for parameters from multiple `Field`, and want to reuse model:\n  \n    PaitBaseModel can be used only for args parameters, it is the most flexible, PaitBaseModel has most of the features of Pydantic. BaseModel, which is not possible with Pydantic.:\n    ```Python\n    from pait.app.starlette import pait\n    from pait.field import Body, Header\n    from pait.model import PaitBaseModel\n\n\n    class TestModel(PaitBaseModel):\n        uid: int = Body.i()\n        content_type: str = Header.i(default=\'Content-Type\')\n\n\n    @pait()\n    async def test(model: PaitBaseModel):\n        return {\'result\': model.dict()}\n    ```\n- TypeHints is Pydantic.BaseModel, mainly used for parameters are derived from the same `Field`, and want to take the model: \n  \n    BaseModel can only be used with kwargs parameters, and the type hints of the parameters must be a class that inherits from `pydantic.BaseModel`, using the example:\n    ````Python\n    from pydantic import BaseModel\n    from pait.app.starlette import pait\n    from pait.field import Body\n    \n    \n    class TestModel(BaseModel):\n        uid: int\n        user_name: str\n    \n    \n    @pait()\n    async def test(model: BaseModel = Body.i()):\n        return {\'result\': model.dict()}\n    ````\n- When TypeHints is not one of the above two cases:\n  \n    can only be used for kwargs parameters and type hints are not the above two cases, if the value is rarely reused, or if you do not want to create a Model, you can consider this approach\n    ```Python\n    from pait.app.starlette import pait\n    from pait.field import Body\n\n\n    @pait()\n    async def test(uid: int = Body.i(), user_name: str = Body.i()):\n        return {\'result\': {\'uid\': uid, \'user_name\': user_name}}\n    ```\n### 1.3.Field\nField will help pait know how to get data from request.\nBefore introducing the function of Field, let’s take a look at the following example. `pait` will obtain the body data of the request according to Field.Body, and obtain the value with the parameter named key. Finally, the parameter is verified and assigned to the uid.\n```Python\nfrom pait.app.starlette import pait\nfrom pait.field import Body\n\n\n@pait()\nasync def demo_post(\n    # get uid from request body data\n    uid: int = Body.i()\n):\n    pass\n```\nThe following example will use a parameter called default.\nSince you can\'t use Content-Type to name the variables in Python, you can only use content_type to name them according to the naming convention of python, so there is no way to get the value directly from the header, so you can set the value of `alias` to Content-Type, and then Pait can get the value of Content-Type in the Header and assign it to the content_type variable.\nAnother example uses `raw_return` and sets it to True. At this time, `Pait` will not use the parameter name `header_dict` as the key to get the data, but directly assign the data of the entire header to the header_dict.\n\n```Python\nfrom pait.app.starlette import pait\nfrom pait.field import Body, Header\n\n\n@pait()\nasync def demo_post(\n    # get uid from request body data\n    uid: int = Body.i(),\n    # get Content-Type from header\n    content_type: str = Header.i(alias=\'Content-Type\'),\n    header_dict: str = Header.i(raw_return=True)\n\n):\n    pass\n```\nThe above only demonstrates the Body and Header of the field, but there are other fields as well::\n- Field.Body   Get the json data of the current request\n- Field.Cookie Get the cookie data of the current request\n- Field.File   Get the file data of the current request, depending on the web framework will return different file object types\n- Field.Form   Get the form data of the current request, if there are multiple duplicate keys, only the first one will be returned \n- Field.Header Get the header data of the current request\n- Field.Path   Get the path data of the current request (e.g. /api/{version}/test, you can get the version data)\n- Field.Query  Get the url parameters of the current request and the corresponding data, if there are multiple duplicate keys, only the first one will be returned \n- Field.MultiQuery Get the url parameter data of the current request, and return the list corresponding to the key\n- Field.MultiForm Get the form data of the current request, return the list corresponding to the key \n\nAll the fields above are inherited from `pydantic.fields.FieldInfo`, most of the parameters here are for api documentation, see for specific usage[pydantic doc](https://pydantic-docs.helpmanual.io/usage/schema/#field-customisation)\n\n\nIn addition there is a field named Depends, he inherits from `object`, he provides the function of dependency injection, he only supports one parameter and the type of function, and the function\'s parameters are written in the same way as the routing function, the following is an example of the use of Depends, through Depends, you can reuse in each function to get the token function:\n\n```Python\nfrom pait.app.starlette import pait\nfrom pait.field import Body, Depends\n\n\ndef demo_depend(uid: str = Body.i(), password: str = Body.i()) -> str:\n    # fake db\n    token: str = db.get_token(uid, password)\n    return token\n\n\n@pait()\nasync def test_depend(token: str = Depends.i(demo_depend)):\n    return {\'token\': token}\n```\n\n### 1.4.requests object\nAfter using `Pait`, the proportion of the number of times the requests object is used will decrease, so `pait` does not return the requests object. If you need the requests object, you can fill in the parameters like `requests: Requests` (you need to use the TypeHints format) , You can get the requests object corresponding to the web framework\n```Python\nfrom starlette.requests import Request\n\n\n@params_verify()\nasync def demo_post(\n    request: Requests,\n    # get uid from request body data\n    uid: int = Body.i()  \n):\n    pass\n```\n\n### 1.5.Exception\n#### 1.5.1Exception Handling\nPait will leave the exception to the user to handle it. Under normal circumstances, pait will only throw the exception of `pydantic` and `PaitBaseException`. The user needs to catch the exception and handle it by himself, for example:\n```Python\nfrom starlette.applications import Starlette\nfrom starlette.requests import Request\nfrom starlette.responses import Response\n\nfrom pait.exceptions import PaitBaseException\nfrom pydantic import ValidationError\n\nasync def api_exception(request: Request, exc: Exception) -> Response:\n    """\n    Handle exception code    \n    """\n\nAPP = Starlette()\nAPP.add_exception_handler(PaitBaseException, api_exception)\nAPP.add_exception_handler(ValidationError, api_exception)\n```\n#### 1.5.2Error Tip\nWhen you use pait incorrectly, pait will indicate in the exception the file path and line number of the function.\n```Bash\n  File "/home/so1n/github/pait/pait/func_param_handle.py", line 101, in set_value_to_kwargs_param\n    f\'File "{inspect.getfile(func_sig.func)}",\'\nPaitBaseException: \'File "/home/so1n/github/pait/example/starlette_example.py", line 28, in demo_post\\n kwargs param:content_type: <class \\\'str\\\'> = Header(key=None, default=None) not found value, try use Header(key={key name})\'\n```\nIf you need more information, can set the log level to debug to get more detailed information\n```Python\nDEBUG:root:\nasync def demo_post(\n    ...\n    content_type: <class \'str\'> = Header(key=None, default=None) <-- error\n    ...\n):\n    pass\n```\n## 2.Document Generation\n> Note: The function is being improved... \n\nIn addition to parameter verification and conversion, pait also provides the ability to output api documentation, which can be configured with simple parameters to output perfect documentation.\n\nNote: Currently only md, json, yaml type documents and openapi documents for json and yaml are supported for output.For the output of md, json, yaml, see \n[doc example](https://github.com/so1n/pait/blob/master/example/api_doc/example_doc)\n\n`pait` will automatically capture the request parameters and url, method and other information of the routing function.\nIn addition, it also supports labeling some relevant information. These labels will only be loaded into the memory when the Python program starts running, and will not affect the performance of the request, as in the following example:\n```Python\nfrom pait.app.starlette import pait\nfrom pait.model import PaitStatus\n\nfrom example.param_verify.model import UserSuccessRespModel, FailRespModel\n\n\n@pait(\n    author=("so1n",),\n    group="user",\n    status=PaitStatus.release,\n    tag=("user", "post"),\n    response_model_list=[UserSuccessRespModel, FailRespModel],\n)\ndef demo() -> None: \n  pass\n```\nParam:\n- author: List of authors who wrote the interface \n- group: The group to which the interface belongs \n- status: The status of the interface, currently only supports several states of `PaitStatus` \n  - default status:\n    - undefined: undefined \n  - in development:  \n    - design: Interface design \n    - dev: Under development and testing \n  - Development completed:  \n    - integration: 联合调试\n    - complete: development completed \n    - test: testing\n  - online:  \n    - release: online\n  - offline:  \n    - abnormal: The interface is abnormal and needs to be offline \n    - maintenance: In maintenance \n    - archive: archive\n    - abandoned: abandoned\n- tag: interface tag \n- response_model_list: return data, Need to inherit from `pait.model.PaitResponseModel`\n\n### 2.1.openapi\nCurrently pait supports most of the functions of openapi, a few unrealized features will be gradually improved through iterations (response-related more complex)\n\nThe openapi module of pait supports the following parameters (more parameters will be provided in the next version):\n- title: openapi\'s title \n- open_api_info: openapi\'s info param  \n- open_api_tag_list: related description of openapi tag \n- open_api_server_list: openapi server list\n- type_: The type of output, optionally json and yaml \n- filename: Output file name, or if empty, output to terminal\n\nThe following is the sample code output from the openapi documentation (modified by the 1.1 code). See [Example code](https://github.com/so1n/pait/tree/master/example/api_doc) and [doc example](https://github.com/so1n/pait/blob/master/example/api_doc/example_doc)\n```Python\nimport uvicorn\n\nfrom starlette.applications import Starlette\nfrom starlette.routing import Route\nfrom starlette.responses import JSONResponse\n\nfrom pait.field import Body\nfrom pait.app.starlette import pait\nfrom pydantic import (\n    BaseModel,\n    conint,\n    constr,\n)\n\n\n# Create a Model based on Pydantic.BaseModel \nclass PydanticModel(BaseModel):\n    uid: conint(gt=10, lt=1000)  # Whether the auto-check type is int, and whether it is greater than or equal to 10 and less than or equal to 1000 \n    user_name: constr(min_length=2, max_length=4)  # Whether the auto-check type is str, and whether the length is greater than or equal to 2, less than or equal to 4\n\n\n\n# Decorating functions with the pait decorator\n@pait()\nasync def demo_post(\n    # pait through the Body () to know the current need to get the value of the body from the request, and assign the value to the model, \n    # and the structure of the model is the above PydanticModel, he will be based on our definition of the field automatically get the value and conversion and judgment\n    model: PydanticModel = Body.i()\n):\n    # Get the corresponding value to return\n    return JSONResponse({\'result\': model.dict()})\n\n\napp = Starlette(\n    routes=[\n        Route(\'/api\', demo_post, methods=[\'POST\']),\n    ]\n)\n\nuvicorn.run(app)\n# --------------------\n\nfrom pait.app import load_app\nfrom pait.api_doc.open_api import PaitOpenApi\n\n\n# Extracting routing information to pait\'s data module\nload_app(app)\n# Generate openapi for routing based on data from the data module\nPaitOpenApi()\n```\n## 3.How to used in other web framework?\nIf the web framework is not supported, which you are using.\nCan be modified sync web framework according to [pait.app.flask](https://github.com/so1n/pait/blob/master/pait/app/flask.py)\n\nCan be modified async web framework according to [pait.app.starlette](https://github.com/so1n/pait/blob/master/pait/app/starlette.py)\n## 4.IDE Support\nWhile pydantic will work well with any IDE out of the box.\n- [PyCharm plugin](https://pydantic-docs.helpmanual.io/pycharm_plugin/)\n- [Mypy plugin](https://pydantic-docs.helpmanual.io/mypy_plugin/)\n\n## 5.Full example\nFor more complete examples, please refer to [example](https://github.com/so1n/pait/tree/master/example)',
    'author': 'So1n',
    'author_email': 'so1n897046026@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/so1n/pait',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import json
from types import CodeType
from typing import Any, Dict, List, Optional

from pydantic.fields import Undefined

from pait.model import PaitCoreModel, PaitStatus

from .base_parse import PaitBaseParse  # type: ignore

json_type_default_value_dict: Dict[str, Any] = {
    "null": None,
    "bool": True,
    "string": "",
    "float": 0.0,
    "integer": 0,
    "object": {},
    "array": [],
}


class PaitMd(PaitBaseParse):
    def __init__(
        self,
        pait_dict: Dict[str, PaitCoreModel],
        title: str = "Pait Doc",
        use_html_details: bool = True,
        filename: Optional[str] = None,
    ):
        self._use_html_details: bool = use_html_details  # some not support markdown in html
        self._title: str = title
        super().__init__(pait_dict)

        markdown_text: str = self.gen_markdown_text()
        self.output(filename, markdown_text, ".md")

    @staticmethod
    def gen_example_json(field_dict_list: List[dict], blank_num: int, indent: int = 2) -> str:
        """
        gen example json by field dict list
        :param field_dict_list:
        :param blank_num: Indent the first character of each line of json
        :param indent: json indent
        :return:
        blank_num: -> .
        indent: -> `
        ....{
        ....``"code": 0,
        ....``"msg": "msg,
        ....}
        """
        blank_num_str: str = " " * blank_num
        example_dict: dict = {}
        for field_dict in field_dict_list:
            sub_dict: dict = example_dict
            key: str = field_dict["param_name"]
            if "." in key:
                key_list: List[str] = key.split(".")
                for sub_key in key_list:
                    key = sub_key
                    if sub_key == key_list[-1]:
                        continue
                    if sub_key not in sub_dict:
                        sub_dict[sub_key] = {}
                    sub_dict = sub_dict[sub_key]
            sub_dict[key] = json_type_default_value_dict[field_dict["type"]]
        json_str: str = f"\n".join([blank_num_str + i for i in json.dumps(example_dict, indent=indent).split("\n")])
        return f"{blank_num_str}```json\n{json_str}\n{blank_num_str}```\n\n"

    @staticmethod
    def gen_md_param_table(field_dict_list: List[dict], blank_num: int = 8) -> str:
        """
        gen param md table
        :param field_dict_list:
        :param blank_num: table str indent
        :return:
            |param name|type|default value|description|other|
            |---|---|---|---|---|
            |age|string|10|user age||
        """
        blank_num_str: str = " " * blank_num
        markdown_text: str = f"{blank_num_str}|param name|type|default value|description|other|\n"
        markdown_text += f"{blank_num_str}|---|---|---|---|---|\n"
        field_dict_list = sorted(field_dict_list, key=lambda x: x["param_name"])
        for field_info_dict in field_dict_list:
            default = field_info_dict["default"]
            if default is Undefined:
                default = "**`Required`**"
            type_ = field_info_dict["type"]
            if default is Undefined:
                type_ = "**`Required`**"
            markdown_text += (
                f"{blank_num_str}"
                f"|{field_info_dict['param_name']}"
                f"|{type_}"
                f"|{default}"
                f"|{field_info_dict['description']}"
                f"|{field_info_dict['other'] or ''}"
                f"|\n"
            )
        return markdown_text

    def gen_markdown_text(self) -> str:
        markdown_text: str = f"# {self._title}\n"
        for group in self._group_list:
            # group
            if self._use_html_details:
                markdown_text += f"<details><summary>Group: {group}</summary>\n\n"
            else:
                markdown_text += f"## Group: {group}\n\n"
            for pait_model in self._group_pait_dict[group]:
                # func info
                markdown_text += f"### Name: {pait_model.operation_id}\n\n"
                status_text: str = ""
                if pait_model.status in (PaitStatus.test, PaitStatus.design, PaitStatus.dev, PaitStatus.integration):
                    status_text = f"<font color=#00BFFF>{pait_model.status.value}</font>"
                elif pait_model.status in (PaitStatus.release, PaitStatus.complete):
                    status_text = f"<font color=#32CD32>{pait_model.status.value}</font>"
                elif pait_model.status in (PaitStatus.abandoned, PaitStatus.abnormal):
                    status_text = f"<font color=#DC143C>{pait_model.status.value}</font>"
                elif pait_model.status:
                    status_text = f"{pait_model.status.value}"

                # func or interface details
                func_code: CodeType = pait_model.func.__code__
                markdown_text += f"|Author|Status|func|description|\n"
                markdown_text += f"|---|---|---|---|\n"
                markdown_text += (
                    f"|{','.join(pait_model.author)}"
                    f"|{status_text}"
                    f'|<abbr title="file:{func_code.co_filename};line: {func_code.co_firstlineno}">'
                    f"{pait_model.func.__qualname__}</abbr>"
                    f"|{pait_model.desc}|\n"
                )

                # request info
                markdown_text += f"- Path: {pait_model.path}\n"
                markdown_text += f"- Method: {','.join(pait_model.method_list)}\n"
                markdown_text += f"- Request:\n"

                field_dict: Dict[str, List[Dict[str, Any]]] = self._parse_func_param_to_field_dict(pait_model.func)
                # request body info
                field_key_list: List[str] = sorted(field_dict.keys())
                for field in field_key_list:
                    field_dict_list = field_dict[field]
                    markdown_text += f"{' ' * 4}- {field.capitalize()} Param\n\n"
                    markdown_text += self.gen_md_param_table(field_dict_list)

                # response info
                markdown_text += f"- Response:\n\n"
                if pait_model.response_model_list:
                    for resp_model_class in pait_model.response_model_list:
                        resp_model = resp_model_class()
                        markdown_text += f"{' ' * 4}- {resp_model.name}\n\n"
                        markdown_text += f"{' ' * 8}|status code|media type|description|\n"
                        markdown_text += f"{' ' * 8}|---|---|---|\n"
                        markdown_text += (
                            f"{' ' * 8}|{','.join([str(i) for i in resp_model.status_code])}"
                            f"|{resp_model.media_type}"
                            f"|{resp_model.description}"
                            f"|\n"
                        )
                        if resp_model.header:
                            markdown_text += f"{' ' * 8}- Header\n"
                            markdown_text += f"{' ' * 12}{resp_model.header}\n"
                        if resp_model.response_data:
                            markdown_text += f"{' ' * 8}- Data\n\n"
                            field_dict_list = self._parse_schema(resp_model.response_data.schema())
                            markdown_text += self.gen_md_param_table(field_dict_list, blank_num=12)
                            markdown_text += f"{' ' * 8}- Example Response Data Json\n\n"
                            markdown_text += self.gen_example_json(field_dict_list, blank_num=12)
                markdown_text += "\n"
            if self._use_html_details:
                markdown_text += "</details>"
        return markdown_text



import inspect
from functools import wraps
from typing import Any, Callable, List, Optional, Tuple, Type

from pait.app.base import BaseAppHelper
from pait.g import pait_data
from pait.model import PaitCoreModel, PaitResponseModel, PaitStatus
from pait.param_handle import async_class_param_handle, async_func_param_handle, class_param_handle, func_param_handle
from pait.util import FuncSig, get_func_sig


def pait(
    app_helper_class: "Type[BaseAppHelper]",
    author: Optional[Tuple[str]] = None,
    desc: Optional[str] = None,
    status: Optional[PaitStatus] = None,
    group: Optional[str] = None,
    tag: Optional[Tuple[str, ...]] = None,
    response_model_list: Optional[List[Type[PaitResponseModel]]] = None,
) -> Callable:
    if not isinstance(app_helper_class, type):
        raise TypeError(f"{app_helper_class} must be class")
    if not issubclass(app_helper_class, BaseAppHelper):
        raise TypeError(f"{app_helper_class} must sub {BaseAppHelper.__class__.__name__}")
    app_name: str = getattr(app_helper_class, "app_name", "")

    def wrapper(func: Callable) -> Callable:
        func_sig: FuncSig = get_func_sig(func)
        qualname = func.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0]

        pait_id: str = f"{qualname}_{id(func)}"
        setattr(func, "_pait_id", pait_id)
        pait_data.register(
            app_name,
            PaitCoreModel(
                author=author,
                desc=desc,
                func=func,
                pait_id=pait_id,
                status=status,
                group=group,
                tag=tag,
                response_model_list=response_model_list,
            ),
        )

        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def dispatch(*args: Any, **kwargs: Any) -> Callable:
                # only use in runtime, support cbv
                class_ = getattr(inspect.getmodule(func), qualname)
                # real param handle
                app_helper: BaseAsyncAppHelper = app_helper_class(class_, args, kwargs)  # type: ignore
                # auto gen param from request
                func_args, func_kwargs = await async_func_param_handle(app_helper, func_sig)
                # support sbv
                await async_class_param_handle(app_helper)
                return await func(*func_args, **func_kwargs)

            return dispatch
        else:

            @wraps(func)
            def dispatch(*args: Any, **kwargs: Any) -> Callable:
                # only use in runtime
                class_ = getattr(inspect.getmodule(func), qualname)
                # real param handle
                app_helper: BaseSyncAppHelper = app_helper_class(class_, args, kwargs)  # type: ignore
                # auto gen param from request
                func_args, func_kwargs = func_param_handle(app_helper, func_sig)
                # support sbv
                class_param_handle(app_helper)
                return func(*func_args, **func_kwargs)

            return dispatch

    return wrapper



from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union, get_type_hints

from pydantic import BaseModel

from pait.util import create_pydantic_model

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, DictStrAny, MappingIntStrAny


class PaitStatus(Enum):
    """Interface life cycle"""

    undefined = "undefined"
    # The interface is under development and will frequently change
    design = "design"
    dev = "dev"

    # The interface has been completed, but there may be some bugs
    integration = "integration"
    complete = "complete"
    test = "test"

    # The interface is online
    release = "release"

    # The interface has been online, but needs to be offline for some reasons
    abnormal = "abnormal"
    maintenance = "maintenance"
    archive = "archive"
    abandoned = "abandoned"


@dataclass()
class PaitResponseModel(object):
    """response model"""

    description: Optional[str] = ""
    header: dict = field(default_factory=dict)
    media_type: str = "application/json"
    response_data: Optional[Type[BaseModel]] = None
    status_code: List[int] = field(default_factory=lambda: [200])

    name: Optional[str] = None

    def __post_init__(self) -> None:
        if not self.name:
            self.name = self.__class__.__name__


class PaitBaseModel(object):
    """pait base model, The value of the attribute supports filling in the pait field
    >>> from pait import field
    >>> class Demo(PaitBaseModel):
    ...     a: int = field.Query.i()
    ...     b: str = field.Body.i()
    """

    _pydantic_model: Optional[Type[BaseModel]] = None

    @classmethod
    def to_pydantic_model(cls) -> Type[BaseModel]:
        if cls._pydantic_model is not None:
            return cls._pydantic_model
        else:
            annotation_dict: Dict[str, Tuple[Type, Any]] = {
                param_name: (annotation, getattr(cls, param_name, ...))
                for param_name, annotation in get_type_hints(cls).items()
                if not param_name.startswith("_")
            }
            cls._pydantic_model = create_pydantic_model(annotation_dict)
            return cls._pydantic_model

    def dict(
        self,
        include: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        exclude: Union["AbstractSetIntStr", "MappingIntStrAny"] = None,
        by_alias: bool = False,
        skip_defaults: bool = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> "DictStrAny":
        if self._pydantic_model is None:
            raise NotImplementedError
        _pydantic_model: BaseModel = self._pydantic_model(**self.__dict__)
        return _pydantic_model.dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

    @classmethod
    def schema(cls, by_alias: bool = True) -> "DictStrAny":
        return cls.to_pydantic_model().schema(by_alias=by_alias)


class PaitCoreModel(object):
    def __init__(
        self,
        func: Callable,
        pait_id: str,
        path: Optional[str] = None,
        method_set: Optional[Set[str]] = None,
        operation_id: Optional[str] = None,
        func_name: Optional[str] = None,
        author: Optional[Tuple[str, ...]] = None,
        desc: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[str, ...]] = None,
        response_model_list: Optional[List[Type[PaitResponseModel]]] = None,
    ):
        self.func: Callable = func  # route func
        self.pait_id: str = pait_id  # qualname + func hash id
        self.path: str = path or ""  # request url path
        self.method_list: List[str] = sorted(list(method_set or set()))  # request method set
        self.operation_id: Optional[str] = operation_id or None  # route name
        self.func_name: str = func_name or func.__name__
        self.author: Tuple[str, ...] = author or ("",)  # The main developer of this func
        self.desc = desc or func.__doc__ or ""  # desc of this func
        self.status: PaitStatus = status or PaitStatus.undefined  # Interface development progress (life cycle)
        self.group: str = group or "root"  # Which group this interface belongs to
        self.tag: Tuple[str, ...] = tag or ("default",)  # Interface tag
        self.response_model_list: Optional[List[Type[PaitResponseModel]]] = response_model_list
